/* Time-stamp: <Tue Dec 13 2005 11:43:58 hwloidl>
   $Id$

   Canonise trees of intermediate code (Chapter 8.1)
   Adapted from on-line version by A. Appel.
*/

import java.util.LinkedList;

class MoveCall extends Stm {
  TEMP dst;
  CALL src;
  MoveCall(TEMP d, CALL s) {dst=d; src=s;}
  public LinkedList<Exp> kids() {return src.kids();}
  public Stm build(LinkedList<Exp> kids) {
	return new MOVE(dst, src.build(kids));
  }
  public <A> A accept (IntmVisitor<A> v) {
        return v.visit (this);
  }
}   
  
class ExpCall extends Stm {
  CALL call;
  ExpCall(CALL c) {call=c;}
  public LinkedList<Exp> kids() {return call.kids();}
  public Stm build(LinkedList<Exp> kids) {
	return new EXP(call.build(kids));
  }
  public <A> A accept (IntmVisitor<A> v) {
        return v.visit (this);
  }
}   

/* (Stm, [Exp]) */  
class StmExpList {
  Stm stm;
  LinkedList<Exp> exps;
  StmExpList(Stm s, LinkedList<Exp> e) {stm=s; exps=e;}
}

public class Canon {
  
 static boolean isNop(Stm a) {
   return a instanceof EXP
          && ((EXP)a).exp instanceof CONST;
 }

 static Stm seq(Stm a, Stm b) {
    if (isNop(a)) return b;
    else if (isNop(b)) return a;
    else return new SEQ(a,b);
 }

 static boolean commute(Stm a, Exp b) {
    return isNop(a)
        || b instanceof NAME
        || b instanceof CONST;
 }

 static Stm do_stm(SEQ s) { 
	return seq(do_stm(s.first), do_stm(s.second));
 }

 static Stm do_stm(MOVE s) { 
	if (s.dest instanceof TEMP 
	     && s.src instanceof CALL) 
		return reorder_stm(new MoveCall((TEMP)s.dest,
						(CALL)s.src));
	else if (s.dest instanceof ESEQ)
	    return do_stm(new SEQ(((ESEQ)s.dest).stm,
					new MOVE(((ESEQ)s.dest).exp,
						  s.src)));
	else return reorder_stm(s);
 }

 static Stm do_stm(EXP s) { 
	if (s.exp instanceof CALL)
	       return reorder_stm(new ExpCall((CALL)s.exp));
	else return reorder_stm(s);
 }

 static Stm do_stm(Stm s) {
     if (s instanceof SEQ) return do_stm((SEQ)s);
     else if (s instanceof MOVE) return do_stm((MOVE)s);
     else if (s instanceof EXP) return do_stm((EXP)s);
     else return reorder_stm(s);
 }

 static Stm reorder_stm(Stm s) {
     StmExpList x = reorder(s.kids());
     return seq(x.stm, s.build(x.exps));
 }

 static ESEQ do_exp(ESEQ e) {
      Stm stms = do_stm(e.stm);
      ESEQ b = do_exp(e.exp);
      return new ESEQ(seq(stms,b.stm), b.exp);
  }

 static ESEQ do_exp (Exp e) {
       if (e instanceof ESEQ) return do_exp((ESEQ)e);
       else return reorder_exp(e);
 }
         
 static ESEQ reorder_exp (Exp e) {
     StmExpList x = reorder(e.kids());
     return new ESEQ(x.stm, e.build(x.exps));
 }

 static StmExpList nopNull = new StmExpList(new EXP(new CONST(0)),null);

 static StmExpList reorder(LinkedList<Exp> exps) {
     if (exps==null || exps.size()==0) {
       return nopNull;
     } else {
       Exp a = Utils.head(exps); // was: .head;
       if (a instanceof CALL) { // bind return value; see p.168
         Temp t = new Temp();
	 Exp e = new ESEQ(new MOVE(new TEMP(t), a),
				    new TEMP(t));
         return reorder(Utils.cons(e,Utils.tail(exps))); // e replaces a!
       } else {
	 ESEQ aa = do_exp(a);
	 StmExpList bb = reorder(Utils.tail(exps));
	 if (commute(bb.stm, aa.exp)) {
	   return new StmExpList(seq(aa.stm,bb.stm), Utils.cons(aa.exp,bb.exps));
	 } else {
	   Temp t = new Temp();
	   return new StmExpList(
			  seq(aa.stm, 
			    seq(new MOVE(new TEMP(t),aa.exp),
				 bb.stm)),
			  Utils.cons(new TEMP(t), bb.exps));
	 }
       }
     }
 }
        
 static LinkedList<Stm> linear(SEQ s, LinkedList<Stm> l) {
      return linear(s.first,linear(s.second,l));
 }
 static LinkedList<Stm> linear(Stm s, LinkedList<Stm> l) {
    if (s instanceof SEQ) {
      return linear((SEQ)s, l);
    } else {
      return Utils.cons(s,l);
    }
 }

 static public LinkedList<Stm> linearize(Stm s) {
    return linear(do_stm(s), null);
 }
}
