/* Time-stamp: <Wed Dec 05 2007 21:14:59 Stardate: [-29]8614.21 hwloidl>
   $Id: IntmTrafo.java,v 1.1 2006/01/16 18:08:22 hwloidl Exp $

   A Compiler for MiniJava: 
     Intm->Intm
   Practical on compiler design, LMU 2007/08
   Based on Appels book on "Modern Compiler Implementation"

   Transformations on the intermediate code. 
   This is the top-level module for the code in Chapter 8.2
   The steps of the transformation are done in modules:
    Canon.java
    BasicBlocks.java
    TraceSchedule.java
   Adapted from on-line version by A. Appel.
*/

import java.util.LinkedList;

public class IntmTrafo {

  public static LinkedList<Stm> flattenCode(LinkedList<LinkedList<Stm>> xs) {
    LinkedList<Stm> res = new LinkedList<Stm>();
    java.util.Iterator<LinkedList<Stm>> it = xs.iterator();
    while (it.hasNext()){
      res.addAll(it.next());
    }
    return res;
  }
			      
  public static LinkedList<Stm> intmTrafo (Stm body) {
    // Step 1: linearise the SEQ-tree into a list of Stms (Canon.java; Ch 8.1)
    LinkedList<Stm> stms = new Canon().linearize(body);
    // System.out.println(".. codeseq after linearize:" + Utils.show_codeseq(stms));
    // Step 2: break the list of Stms into a list of basic blocks (BasicBlocks.java; Ch 8.2)
    BasicBlocks bbs = new BasicBlocks(stms);
    // System.out.println(".. basic blocks:" + Utils.show_codeseq(bbs.getBlocks()));
    // Step 3: generate traces out of BBS, doing some reordering of BBs (TraceSchedule.java; Ch 8.3)
    TraceSchedule trs = new TraceSchedule(bbs); // NB: result is in .newBlocks
    return flattenCode(trs.newBlocks);          // flatten list of BBs into list of Stms
  }
}
