/* Time-stamp: <Wed Dec 05 2007 15:00:08 Stardate: [-29]8612.91 hwloidl>
   $Id$

   Generate traces out of a list of basic blocks (BBs) (Chapter 8.3)
   Result is left in: theBlocks
   Adapted from on-line version by A. Appel.
*/


import java.util.LinkedList;

public class TraceSchedule {

  public LinkedList<Stm> stms;
  BasicBlocks theBlocks;
  public LinkedList<LinkedList<Stm>> newBlocks = new LinkedList<LinkedList<Stm>>();

  // this table maps labels to BBs; initially all BBs are in the table
  // after having been processed they are `marked' i.e. removed from the table
  java.util.HashMap<Label,LinkedList<Stm>> unmarkedBBs = new java.util.HashMap();

  private int count = 0;

  // start with BB l and generate a trace; l modified along the way
  private LinkedList<Stm> trace(LinkedList<Stm> l) {
    LinkedList<Stm> currBlock = new LinkedList<Stm>(); //start of trace; l will be moved!
    currBlock.addAll(l);
    LABEL lab = (LABEL)Utils.head(l);
    //System.out.println("++ starting trace from label " + lab.label);
    for(;;) {                               // iteration var is l: current BB
      // System.out.println("++ trace_loop: currBlock is now:" + Utils.show_codeseq(currBlock));
     unmarkedBBs.remove(lab.label);         // mark this BB
     Stm s = currBlock.getLast();           // last Stm in BB
     if (s instanceof JUMP) {               // last statement of the BB is a jump
       JUMP j = (JUMP)s;                    // pick first of all possible targets of jump
       LinkedList<Stm> target = (LinkedList<Stm>)unmarkedBBs.get(Utils.head(Utils.list2JList(j.poss)));
       // if only one target, which is not marked yet, add the entire BB to current trace
       //ngoq ngo': if (Utils.tail(Utils.list2JList(j.poss))==null && target!=null) { 
       if (j.poss.len()<=1) { 
	 if (target!=null) {
	   // NEVER MERGE: currBlock.removeLast();   // remove last JUMP instr (now obsolete)
	   currBlock.addAll(target); // and merge this BB with target BB
	   lab=(LABEL)Utils.head(target);
	   // System.out.println("++ merging BB with label " + lab.label);
	   // if (l==null) { System.out.println("PANIC: TraceSchedule.trace: l==null"); }
	 } else { // the only target has been processed already -> finished
	   return currBlock;
	 }
       } else { // more than one target -> add next target to the end of the trace
	 if (target!=null) {
	   lab=(LABEL)Utils.head(target);
	   // System.out.println("++ several poss targets " + (new Integer(j.poss.len())).toString() + " or target " + ((LABEL)Utils.head(target)).label + " marked already ; adding next BB to currBlock without a merge ");
	   currBlock.addAll(getNext(lab.label)); // no merge in this case
	   return currBlock;
	 } else { // the only target has been processed already -> finished
	   return currBlock;
	 }
       }
     } else if (s instanceof CJUMP) {       // last statement of BB is a CJUMP
	CJUMP j = (CJUMP)s;
        LinkedList<Stm> t = (LinkedList<Stm>)unmarkedBBs.get(j.trueLab);
        LinkedList<Stm> f = (LinkedList<Stm>)unmarkedBBs.get(j.falseLab);
        if (f!=null) {                      // unmarked false branch
	  currBlock.addAll(f);              // add it to trace now (right after CJUMP!)
	  lab=(LABEL)Utils.head(f);
	  // System.out.println("++ CJUMP fallthrough with false label " + lab.label);
	} else if (t!=null) {    // replace last CJUMP by a CJUMP with swapped branches
	  currBlock.removeLast();
	  currBlock.addLast(new CJUMP(CJUMP.notRel(j.rel),
					j.left,j.right,
					j.falseLab,j.trueLab));
	  currBlock.addAll(t);        // add old true branch, which is now false branch
	  lab=(LABEL)Utils.head(t);
	  // System.out.println("++ CJUMP swap branches with true label " + lab.label);
        } else {                 // both branches are marked (see p. 172)
	  currBlock.removeLast();
	  Label ff = new Label(); // invent new label to make CJUMP followed by false lab
	  Label lab_rest = labelOfNext();
	  LinkedList<Stm> rest;
	  if (lab_rest!=null) {
	    rest = getNext(lab_rest);
	  } else {
	    rest = new LinkedList<Stm>();
	  }
	  currBlock.addLast(new CJUMP(j.rel,j.left,j.right,
					j.trueLab,ff));
	  currBlock.addAll(Utils.cons(new LABEL(ff),
		          Utils.cons(new JUMP(j.falseLab),
					       rest)));
	  // System.out.println("++ CJUMP dummy label " + ff.toString() + " added to curr block " + lab.label);
	  return currBlock;
        }
     } else {                    // not a JUMP or CJUMP at the end of a BB
       throw new Error("Bad basic block in TraceSchedule");
     }
    }
  }

  private void addTrace(LinkedList<Stm> bb) {
      newBlocks.addLast(bb);
      count++;
      // System.out.println("__ TraceSchedule.getNext(): added final label as block no. " + (new Integer(count)).toString() + Utils.show_codeseq(bb));
  }
    
  // main loop iterating over theBlocks.blocks
  void iterate_getNext() {
    Label labOfNextBB;
    for (labOfNextBB=labelOfNext(); labOfNextBB!=null; labOfNextBB=labelOfNext()) { 
      // System.out.println("__ TraceSchedule.iterate_getNext(): next BB to process is " + labOfNextBB.toString() + " ; then " + (new Integer(theBlocks.blocks.size())).toString() + " more blocks in queue");
      addTrace(getNext(labOfNextBB));
    }
    // insert final label <done>
    LinkedList<Stm> bb = new LinkedList<Stm>();
    bb.addLast((Stm)new LABEL(theBlocks.done));
    addTrace(bb);
    // System.out.println("__ TraceSchedule.iterate_getNext(): added done label " + theBlocks.done + " as last trace");    
    return ;
  }

  private Label labelOfNext() {
    if (! (theBlocks.blocks==null || theBlocks.blocks.size()==0)) {
      LinkedList<Stm> bb = Utils.head(theBlocks.blocks);
      if (Utils.head(bb) instanceof LABEL) {
	return ((LABEL)Utils.head(bb)).label;
      } else {
	System.out.println("PANIC: TraceSchedule.labelOfNext: BB does not start with a label");
	return null;
      }
    } else {
      return null;
    }
  }

  // get next trace
  LinkedList<Stm> getNext(Label lab) {
    LinkedList<Stm> bb = unmarkedBBs.get(lab);
      if (bb != null) { // unmarked -> trace it
	return trace(bb);
      } else { // marked -> remove it and process next BB
	// System.out.println("__ TraceSchedule.getNext(): nothing to do for BB " + lab);
	theBlocks.blocks.removeFirst(); //  = theBlocks.blocks.tail;
	return new LinkedList<Stm>();
      }
  }
  

  // modified the list of blocks in b.blocks into a list of traces
  public TraceSchedule(BasicBlocks b) { 
    theBlocks=b;
    // System.out.println("__ starting TraceSchedule with that many blocks to process: " + (new Integer(theBlocks.blocks.size())).toString());
    // l iterates over the BBs in the list b
    java.util.Iterator<LinkedList<Stm>> it = b.blocks.iterator();
    // add the label of each BB to the Map <unmarkedBBs> of BBs still to be processed
    while(it.hasNext()) {
      LinkedList<Stm> bb = it.next();
      unmarkedBBs.put(((LABEL)Utils.head(bb)).label, bb);
    }
    
    iterate_getNext();
    unmarkedBBs=null;
  }        
}


